@@ -15,6 +15,7 @@
 #include <climits>
 #include <cstddef>
 #include <cstdint>
+#include <cstring>
 #include <istream>
 #include <ostream>
 #include <string>
@@ -22,6 +23,7 @@
 
 #include "absl/log/absl_check.h"
 #include "absl/log/absl_log.h"
+#include "absl/log/log.h"
 #include "absl/strings/cord.h"
 #include "absl/strings/cord_buffer.h"
 #include "absl/strings/internal/resize_uninitialized.h"
@@ -45,15 +47,23 @@
 namespace google {
 namespace protobuf {
 
-void MessageLite::DestroyInstance(bool free_memory) {
+void MessageLite::DestroyInstance() {
 #if defined(PROTOBUF_CUSTOM_VTABLE)
-  _class_data_->delete_message(this, free_memory);
+  _class_data_->destroy_message(*this);
 #else   // PROTOBUF_CUSTOM_VTABLE
-  ABSL_DCHECK(!free_memory);
   this->~MessageLite();
 #endif  // PROTOBUF_CUSTOM_VTABLE
 }
 
+void MessageLite::DeleteInstance() {
+  // Cache the size and pointer because we can't access them after the
+  // destruction.
+  const size_t size = GetClassData()->allocation_size();
+  void* const ptr = this;
+  DestroyInstance();
+  internal::SizedDelete(ptr, size);
+}
+
 void MessageLite::CheckTypeAndMergeFrom(const MessageLite& other) {
   auto* data = GetClassData();
   auto* other_data = other.GetClassData();
@@ -64,23 +74,15 @@ void MessageLite::CheckTypeAndMergeFrom(const MessageLite& other) {
   data->merge_to_from(*this, other);
 }
 
-#if defined(PROTOBUF_CUSTOM_VTABLE)
-uint8_t* MessageLite::_InternalSerialize(
-    uint8_t* ptr, io::EpsCopyOutputStream* stream) const {
-  return _class_data_->serialize(*this, ptr, stream);
-}
-
 MessageLite* MessageLite::New(Arena* arena) const {
-  return static_cast<MessageLite*>(_class_data_->new_message(this, arena));
+  auto* data = GetClassData();
+  // The `instance->New()` expression requires using the actual instance
+  // instead of the prototype for the inner function call.
+  // Certain custom instances have special per-instance state that needs to be
+  // copied.
+  return data->message_creator.New(this, data->prototype, arena);
 }
 
-void MessageLite::Clear() { _class_data_->clear(*this); }
-
-size_t MessageLite::ByteSizeLong() const {
-  return _class_data_->byte_size_long(*this);
-}
-#endif  // PROTOBUF_CUSTOM_VTABLE
-
 bool MessageLite::IsInitialized() const {
   auto* data = GetClassData();
   return data->is_initialized != nullptr ? data->is_initialized(*this) : true;
@@ -91,8 +93,8 @@ const char* MessageLite::_InternalParse(const char* ptr,
   return internal::TcParser::ParseLoop(this, ptr, ctx, GetTcParseTable());
 }
 
-std::string MessageLite::GetTypeName() const {
-  return std::string(TypeId::Get(*this).name());
+internal::GetTypeNameReturnType MessageLite::GetTypeName() const {
+  return internal::GetTypeNameReturnType(TypeId::Get(*this).name());
 }
 
 absl::string_view TypeId::name() const {
@@ -102,7 +104,7 @@ absl::string_view TypeId::name() const {
   }
 
   // For LITE messages, the type name is a char[] just beyond ClassData.
-  return reinterpret_cast<const char*>(data_) + sizeof(MessageLite::ClassData);
+  return reinterpret_cast<const char*>(data_) + sizeof(internal::ClassData);
 }
 
 void MessageLite::OnDemandRegisterArenaDtor(Arena* arena) {
@@ -137,15 +139,9 @@ std::string MessageLite::DebugString() const {
   return absl::StrCat("MessageLite at 0x", absl::Hex(this));
 }
 
+#if !defined(PROTOBUF_CUSTOM_VTABLE)
 int MessageLite::GetCachedSize() const { return AccessCachedSize().Get(); }
-
-internal::CachedSize& MessageLite::AccessCachedSize() const {
-  auto* data = GetClassData();
-  ABSL_DCHECK(data != nullptr);
-  ABSL_DCHECK(data->cached_size_offset != 0);
-  return *reinterpret_cast<internal::CachedSize*>(const_cast<char*>(
-      reinterpret_cast<const char*>(this) + data->cached_size_offset));
-}
+#endif
 
 namespace {
 
@@ -653,7 +649,7 @@ bool MessageLite::AppendPartialToCord(absl::Cord* output) const {
   const size_t size = ByteSizeLong();
   const size_t total_size = size + output->size();
   if (size > INT_MAX) {
-    ABSL_LOG(ERROR) << "Exceeded maximum protobuf size of 2GB.";
+    ABSL_LOG(ERROR) << "Exceeded maximum protobuf size of 2GB: " << size;
     return false;
   }
 
